/* Copyright 2008-2014 Research Foundation State University of New York   */

/* This file is part of QUB Express.                                      */

/* QUB Express is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Express is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

/*
begin_html
<b><a href="index.html">Up</a></b>
<ul>
<li><a href="qubfast.h.html">qubfast.h</a></li>
</ul>
end_html
*/

#ifdef _WIN32
  #include <windows.h>
  #include <time.h>
  #define sleep Sleep
#else
  #include <stdlib.h>
  #define BOOL int
  #define TRUE 1
  #define FALSE 0
#endif

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <math.h>
#include <string.h>
#include <vector>
using std::vector;
#include <algorithm>
using std::sort;

#include "qub_idlstats.h"

template<class T>
T median(std::vector<T>& lst) {
	sort(lst.begin(), lst.end());
	int h = ((int) lst.size()) / 2;
	if ( lst.size() == 0 )
		return 0.0;
	else if ( lst.size() % 2 )
		return lst[h];
	else
		return T((lst[h] + lst[h-1]) / 2.0);
}

class qub_idlstats
{
public:
  double sampling;
  vector<double> means, stds, lifetimes, median_lifetimes, occupancies;
  vector<int> nevents, npoints;
  vector< vector<double> > lifetime_lists;
  double total_dur;
  int last_cls;
  
  qub_idlstats(double sampling_)
    : sampling(sampling_), total_dur(0.0), last_cls(-1)
  {}
  
  void reset()
  {
    means.resize(0);
    stds.resize(0);
    lifetimes.resize(0);
    median_lifetimes.resize(0);
    occupancies.resize(0);
    nevents.resize(0);
    npoints.resize(0);
    total_dur = 0.0;
    last_cls = -1;
  }
  void add(float *data, int ndwell, int offset, int *firsts, int *lasts, int *classes)
  {
    if ( ! stds.size() ) { // first pass
      for (int i=0; i<ndwell; ++i) {
	int c = classes[i];
	if ( c < 0 ) {
	  last_cls = c;
	  continue;
	}
	if ( c >= (int) means.size() ) {
	  means.resize(c+1);
	  lifetimes.resize(c+1);
	  lifetime_lists.resize(c+1);
	  median_lifetimes.resize(c+1);
	  occupancies.resize(c+1);
	  nevents.resize(c+1);
	  npoints.resize(c+1);
	}
	int n = lasts[i] - firsts[i] + 1;
	double dur = sampling * n;
	if ( c == last_cls ) {
	  lifetime_lists[c][lifetime_lists[c].size()-1] += dur;
	} else {
	  lifetime_lists[c].push_back(dur);
	  nevents[c] += 1;
	  last_cls = c;
	}
	lifetimes[c] += dur;
	total_dur += dur;
	npoints[c] += n;
	float *dwdat = data + (firsts[i] - offset);
	double sum = 0.0;
	for (int p=0; p<n; ++p)
	  sum += dwdat[p];
	means[c] += sum;
      }
    } else { // std pass
      for (int i=0; i<ndwell; ++i) {
	int c = classes[i];
	if ( c < 0 )
	  continue;
	if ( c >= (int) stds.size() )
	  stds.resize(c+1);
	int n = lasts[i] - firsts[i] + 1;
	float *dwdat = data + (firsts[i] - offset);
	double sum = 0.0;
	for (int p=0; p<n; ++p) {
	  double diff = dwdat[p] - means[c];
	  sum += diff * diff;
	}
	stds[c] += sum;
      }
    }
  }
  int add_up_means()
  {
    int Nclass = (int) means.size();
    
    for (int c=0; c<Nclass; ++c) {
      occupancies[c] = total_dur ? (lifetimes[c] / total_dur) : 0.0;
      lifetimes[c] = nevents[c] ? (lifetimes[c] / nevents[c]) : 0.0;
      median_lifetimes[c] = median(lifetime_lists[c]);
      means[c] = npoints[c] ? (means[c] / npoints[c]) : 0.0;
    }
    
    stds.resize(Nclass);
    return Nclass;
  }
  int add_up_stds()
  {
    int Nclass = (int) stds.size();
    
    for (int c=0; c<Nclass; ++c)
      stds[c] = npoints[c] ? sqrt(stds[c] / npoints[c]) : 0.0;
    
    return Nclass;
  }
};

extern "C"  QUBFAST_API void*    qub_idlstats_create(double sampling)
{
  return new qub_idlstats(sampling);
}

extern "C"  QUBFAST_API void     qub_idlstats_free(void *idls)
{
  delete (qub_idlstats *) idls;
}

extern "C"  QUBFAST_API void     qub_idlstats_reset(void *idls)
{
  qub_idlstats *is = (qub_idlstats *) idls;
  is->reset();
}

extern "C"  QUBFAST_API void     qub_idlstats_add(void *idls, float *data, int ndwell, int offset,
						  int *firsts, int *lasts, int *classes)
{
  qub_idlstats *is = (qub_idlstats *) idls;
  is->add(data, ndwell, offset, firsts, lasts, classes);
}

extern "C"  QUBFAST_API int      qub_idlstats_add_up_means(void *idls) // returns Nclass
{
  qub_idlstats *is = (qub_idlstats *) idls;
  return is->add_up_means();
}

extern "C"  QUBFAST_API double*  qub_idlstats_get_means(void *idls)
{
  qub_idlstats *is = (qub_idlstats *) idls;
  return & is->means[0];
}

extern "C"  QUBFAST_API double*  qub_idlstats_get_lifetimes(void *idls)
{
  qub_idlstats *is = (qub_idlstats *) idls;
  return & is->lifetimes[0];
}

extern "C"  QUBFAST_API double*  qub_idlstats_get_median_lifetimes(void *idls)
{
  qub_idlstats *is = (qub_idlstats *) idls;
  return & is->median_lifetimes[0];
}

extern "C"  QUBFAST_API double*  qub_idlstats_get_occupancies(void *idls)
{
  qub_idlstats *is = (qub_idlstats *) idls;
  return & is->occupancies[0];
}

extern "C"  QUBFAST_API int*     qub_idlstats_get_nevents(void *idls)
{
  qub_idlstats *is = (qub_idlstats *) idls;
  return & is->nevents[0];
}

  // now you re-add it all, then
extern "C"  QUBFAST_API int      qub_idlstats_add_up_stds(void *idls) // returns Nclass
{
  qub_idlstats *is = (qub_idlstats *) idls;
  return is->add_up_stds();
}

extern "C"  QUBFAST_API double*  qub_idlstats_get_stds(void *idls)
{
  qub_idlstats *is = (qub_idlstats *) idls;
  return & is->stds[0];
}


